#ifndef GRAH_INTERFACE_HPP_
#define GRAH_INTERFACE_HPP_

#include "config.h"
#include "genes.hpp"
#include "constants.h"
#include <string>
#include <stdlib.h>
#include <boost/preprocessor.hpp>

namespace allinone {

  class Graph : public Genes{

    private:
    
    bool is_directed;
    bool is_weighted;
    _PFloat threshold;
    
    protected:
    virtual void init(long vector_length, long no_vertices){
      
    }

    public:

    enum Graph_type {
         Graph_none = 0x0, 
         Graph_matrix, 
         Graph_adjacency_list, 
         Graph_adjacency_matrix};

    enum Graph_datatype{
#ifdef GRAPH_INTERFACE_H_
//  #define DATATYPES_TOENUM(r, x, t) x ## t 
#define DATATYPES_HELPER(x, t) (x ## t) 
#define DATATYPES_HELPER2(x, t) x ## t 
#define DATATYPES_TOENUM(r, x, t) DATATYPES_HELPER(x, t) 
      BOOST_PP_SEQ_ENUM(
        BOOST_PP_SEQ_FOR_EACH(DATATYPES_TOENUM, datatype_, (none=0x0)DATATYPES) 
      )
#endif
    } ;
 
    private:
    Graph_type graph_type;
    Graph_datatype graph_datatype;

    public:
    template<class _P>
    static Graph_datatype get_datatype();

    Graph_type get_graphtype()  {return graph_type;}
    Graph_datatype get_datatype()  {return graph_datatype;}
    static string get_datatype_s(Graph_datatype graph_datatype){
      static const string gds[] = {
#define STRINGIFY3(r, x, t) (STRINGIFY(t))
        BOOST_PP_SEQ_ENUM(BOOST_PP_SEQ_FOR_EACH(STRINGIFY3, ~, (none)DATATYPES))
      };
      return gds[graph_datatype];
    }

    static string get_graphtype_s(Graph_type graph_type){
      static const string gts[] = {
        "none", 
        "matrix", 
        "adjacency_list", 
        "adjacency_matrix"
      };
      return gts[graph_type];
    }

    string get_datatype_s(){
      return Graph::get_datatype_s(graph_datatype);
    }

    string get_graphtype_s(){
      return Graph::get_graphtype_s(graph_type);
    }

    Graph(bool is_directed, bool is_weighted, long vector_length, long no_vertices,
        _PFloat threshold, Graph_type graph_type, Graph_datatype graph_datatype) : 
       Genes(no_vertices, InterfaceFlow::Graph), 
       graph_type(graph_type), graph_datatype(graph_datatype),
       is_directed(is_directed), is_weighted(is_weighted), threshold(threshold) {
      init(vector_length, no_vertices); 
    }

    virtual ~Graph(){
    }

    virtual long get_vector_length() const = 0;
    virtual long get_no_vertices()   const = 0;
  
    virtual _PFloat get(int from, int to, int vector_no) const = 0;

    virtual void set(int from, int to, int vector_no, _PFloat v) = 0;
    virtual void set(int from, int to, int vector_no, int v) = 0;
    virtual bool is_edge(int from, int to, int vector_no) const = 0;

    virtual void * graph() {return this;}

    bool get_is_directed() const {return is_directed;}
    bool get_is_weighted() const {return is_weighted;}
    _PFloat get_threshold() const {return threshold;}

    inline _PFloat get(const char * from, const char * to, int vector_no) {
      return get(this->get_no(from), this->get_no(to), vector_no);
    }

    inline void set(const char * from, const char * to, int vector_no, int v){
      set(this->get_no(from), this->get_no(to), vector_no, v);
    }

    inline void set(const char * from, const char * to, int vector_no, _PFloat v){
      set(this->get_no(from), this->get_no(to), vector_no, v);
    }

    virtual inline bool is_edge(const char * from, const char * to, int vector_no){
      return is_edge(this->get_no(from), this->get_no(to), vector_no);
    }

    void set_edges_from_graph(const Graph * _g){
      int iend = _g->get_no_vertices(); 
      for(int i = 0 ; i < iend; ++i)
        for(int j = 0; j < iend; ++j)
          for(int k = 0; k < this->get_vector_length(); ++k){
            if(_g->is_edge(i,j,k))
              this->set(i,j,k, _g->get(i,j,k));
          }
    }
    
  };

#define GRAPH_DATATYPE_GEN(r, x, t)                                           \
  template<> inline Graph::Graph_datatype Graph::get_datatype<t>()            \
      { return DATATYPES_HELPER(x, t); }

  BOOST_PP_SEQ_FOR_EACH(GRAPH_DATATYPE_GEN, datatype_, DATATYPES) 

}
#endif /*GRAH_INTERFACE_HPP_*/

