#ifndef RESULTS_H_
#define RESULTS_H_

#include"config.h"
#include"constants.h"
#include"genes.hpp"
#include<string>
#include<deque>
#include<map>

using namespace std;
  
namespace allinone{

  

  class Results : public Genes{
    public:
    typedef pair<int, _PFloat> ResultBinEl;
    typedef map<int, _PFloat> ResultBin;
    typedef pair<_PFloat, ResultBin> Result;
   
    deque<Result> res;
    protected:
    long no_vertices;
    _PFloat score;

    virtual void init(){
    }

    public:
    Results(_PFloat score, int no_vertices) : 
      Genes(no_vertices, InterfaceFlow::Results),
      no_vertices(no_vertices), score(score) {
      init();
    }

    long get_no_vertices() const {return no_vertices;}
    long get_no_results()  const {return res.size();}
    _PFloat get_score()    const {return score;}
    
    Result & operator[](int i){ 
      this->get(i); 
    }
  
    Result & get(const char *str){
      return get(this->get_no(str));
    }

    Result & get(int i){
#ifdef __DEBUG__
      if(i >= this->get_no_results())
        LOG(WARNING, "%d is out of range; Number of results: %d", i, 
                                              this->get_no_results());
#endif
      return res[i]; 
    }

    deque<Result> & get_deq(){  
      return res;
    }

    typedef deque<Result>::iterator iterator;
    typedef deque<Result>::reverse_iterator reverse_iterator;

    iterator begin(){ return res.begin(); }
    iterator end(){ return res.end(); }
    reverse_iterator rbegin(){ return res.rbegin(); }
    reverse_iterator rend(){ return res.rend(); }
  };

  struct Results_less{
    int reverse;
    Results_less(int reverse = 0) : reverse(reverse){ }

    bool operator() (const Results & _1, const Results & _2) const {
      if (reverse) return _1.get_score() > _2.get_score();
      return _1.get_score() < _2.get_score();
    }

    bool operator() (const Results * _1, const Results * _2) const {
      if (reverse) return _1->get_score() > _2->get_score();
      return _1->get_score() < _2->get_score();
    }
  };
}


#endif /*RESULTS_H_*/
