#ifndef GRAH_BOOST_HPP_
#define GRAH_BOOST_HPP_

#include "config.h"
#include "constants.h"
#include "graph_interface.hpp"
#include <stdlib.h>
#include <string>
#include <assert.h>
#include <string.h>
#include <vector>
#include <map>
#include <cmath>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/adjacency_matrix.hpp>
#include <boost/preprocessor/repetition.hpp>

using std::vector;
using std::map;
using std::string;

namespace graph {

//  using namespace boost;

  template <class _P, long _vector_length>
  struct Edge_property {
    union{
    _P p;
    _P ptr[_vector_length];
    };
//      Edge_property(int n){
//        ptr = new _P[n];
//      }
//      ~Edge_property(){
//        delete []ptr;
//      }
  };

  template<class _P, class G, bool _is_directed, int _vector_length>
  class Graph_boost : public allinone::Graph, G{
  
    private:
    G * g;
    long no_vertices;
    typedef G Graph;
//    Graph *g;

    virtual void init (long no_vertices){
//      g = new G(no_vertices);
      g=this;
    }

    public:
    Graph_boost(bool is_weighted, long no_vertices, Graph_type gt, 
      _P threshold=numeric_limits<_P>::max()) : 
        allinone::Graph(_is_directed, is_weighted, _vector_length, 
        no_vertices, threshold, gt, 
        allinone::Graph::get_datatype<_P>()), G(no_vertices), 
        no_vertices(no_vertices) {
      init(no_vertices); 
    }

    Graph_boost(allinone::Graph * _g, long no_vertices, Graph_type gt, bool copy_data = true) : 
        allinone::Graph(_g->get_is_directed(), _g->get_is_weighted(), 
        _g->get_vector_length(), no_vertices, _g->get_threshold(), gt,
        allinone::Graph::get_datatype<_P>()), G(no_vertices),
        no_vertices(no_vertices)
    { 
      LOG(DEBUG, "Constructor");
//      *this = *_g;
      if(no_vertices < this->no_vertices){
        exit(-1);
      }
      init(no_vertices);
      
      int iend = _g->get_no_vertices(); 
      this->set_names(&_g->get_name(0), iend);
      this->set_descs(&_g->get_desc(0), iend);
       
      if(copy_data)
        this->set_edges_from_graph(_g);
    }

    virtual ~Graph_boost(){
//      delete g;
    }

    long get_vector_length() const {return _vector_length;}
    long get_no_vertices()   const {return no_vertices;}
    
    inline _PFloat get(int from, int to, int vector_no) const{
      _P * ptr = ((Graph_boost *) this)->get_t(from, to, vector_no);
      if(ptr) return *ptr;
      return 0;
    }

    inline _P * get_t(int from, int to, int vector_no){
#ifdef __DEBUG__
      if(min(from, to) < 0 || max(from, to) >= this->no_vertices || 
            vector_no < 0 || vector_no >= _vector_length){ 
        LOG(DEBUG, "Trying to access a vertice[%d][%d][%d]"
                   " which is not in the graph[%d][%d][%d]", 
                   from, to, vector_no, 
                   this->no_vertices, this->no_vertices, _vector_length);
      }
#endif /*__DEBUG__*/
      std::pair<typename Graph::edge_descriptor, bool> ep =
        boost::edge(boost::vertex(from, *g), boost::vertex(to, *g), *g);
      if(!ep.second) return 0;
      return &(*g)[ep.first].ptr[vector_no];
    } 

    template<class T>
    void set_t(int from, int to, int vector_no, T v){
#ifdef __DEBUG__
      if(min(from, to) < 0 || max(from, to) >= this->no_vertices || 
            vector_no < 0 || vector_no >= _vector_length){ 
        LOG(DEBUG, "Trying to access a vertice[%d][%d][%d]"
                   " which is not in the graph[%d][%d][%d]", 
                   from, to, vector_no, 
                   this->no_vertices, this->no_vertices, _vector_length);
      }
#endif /*__DEBUG__*/
      typename Graph::vertex_descriptor vfrom = boost::vertex(from, *g), 
                                          vto = boost::vertex(to, *g);
      std::pair<typename Graph::edge_descriptor, bool> ep =
        boost::edge(vfrom, vto, *g);
      if(!ep.second){
        if(v == 0) return;
        ep = boost::add_edge(vfrom, vto, *g);
      }
      (*g)[ep.first].ptr[vector_no] = v;
    }

    void * graph() { 
      return g; 
    }

    inline void set(int from, int to, int vector_no, _PFloat v){
      set_t(from, to, vector_no, v);
    }

    inline void set(int from, int to, int vector_no, int v){
      set_t(from, to, vector_no, v);
    }


    inline bool is_edge(int from, int to, int vector_no) const{
#ifdef __DEBUG__
      if(min(from, to) < 0 || max(from, to) >= this->no_vertices || 
            vector_no < 0 || vector_no >= _vector_length){ 
        LOG(DEBUG, "Trying to access a vertice[%d][%d][%d]"
                   " which is not in the graph[%d][%d][%d]", 
                   from, to, vector_no, 
                   this->no_vertices, this->no_vertices, _vector_length);
      }
#endif /*__DEBUG__*/
      std::pair<typename Graph::edge_descriptor, bool> ep =
        boost::edge(boost::vertex(from, *g), boost::vertex(to, *g), *g);
      return ep.second && std::abs((*g)[ep.first].ptr[vector_no])
              < this->get_threshold();
    }
   
  };
  
 
#define MAX_VECTOR_LENGTH 1
#define GRAPH_CREATOR(z, n, data)                                             \
    case (n + 1):                                                             \
    if(is_directed)                                                           \
      return new Graph_boost<_P, BASE boost::directedS, boost::no_property,   \
                  Edge_property<_P, n + 1> >, true, n + 1> data;              \
    else                                                                      \
      return new Graph_boost<_P, BASE boost::undirectedS,boost::no_property,  \
                  Edge_property<_P, n + 1> >,false, n + 1> data;              \
    break;


#define BASE boost::adjacency_list<boost::mapS,boost::vecS,

  template<class _P>
  inline allinone::Graph * make_boost_adjacency_list(bool is_directed, 
    bool is_weighted, long vector_length, long no_vertices, 
    _P threshold=numeric_limits<_P>::max()){
    LOG(DEBUG, "Allocating adjacency_list");
    LOG(DEBUG, " using %s", is_directed? "directed" : "undirected");
    LOG(DEBUG, " using type: %s", 
      allinone::Graph::get_datatype_s(allinone::Graph::get_datatype<_P>()).c_str());
    switch (vector_length){
      BOOST_PP_REPEAT(MAX_VECTOR_LENGTH, GRAPH_CREATOR, 
        (true, no_vertices, allinone::Graph::Graph_adjacency_list))
    default:
LOG(ERROR, "Not supported size of vector: %ld; should be [1:%d]", 
      vector_length, MAX_VECTOR_LENGTH);
    }
    return 0;
  }

  template<class _P>
  inline allinone::Graph * make_boost_adjacency_list(allinone::Graph * _g, 
                                  long no_vertices, bool copy_data = true){
    LOG(DEBUG, "Allocating adjacency_list");
    bool is_directed = _g->get_is_directed();
    LOG(DEBUG, " using %s", is_directed? "directed" : "undirected");
    LOG(DEBUG, " using type: %s", 
      allinone::Graph::get_datatype_s(allinone::Graph::get_datatype<_P>()).c_str());
    switch (_g->get_vector_length()){
      BOOST_PP_REPEAT(MAX_VECTOR_LENGTH, GRAPH_CREATOR, 
        (_g, no_vertices, allinone::Graph::Graph_adjacency_list, copy_data))
    default:
LOG(ERROR, "Not supported size of vector: %ld; should be [1:%d]", 
      _g->get_vector_length(), MAX_VECTOR_LENGTH);
    }
    return 0;
  }

#undef BASE 
#define BASE boost::adjacency_matrix<

  template<class _P>
  inline allinone::Graph * make_boost_adjacency_matrix(bool is_directed, 
    bool is_weighted, long vector_length, long no_vertices, 
    _P threshold=numeric_limits<_P>::max()){
    LOG(DEBUG, "Allocating adjacency_matrix");
    LOG(DEBUG, " using %s", is_directed? "directed" : "undirected");
    LOG(DEBUG, " using type: %s", 
      allinone::Graph::get_datatype_s(allinone::Graph::get_datatype<_P>()).c_str());
    switch (vector_length){
      BOOST_PP_REPEAT(MAX_VECTOR_LENGTH, GRAPH_CREATOR, 
        (true, no_vertices, allinone::Graph::Graph_adjacency_matrix))
    default:
LOG(ERROR, "Not supported size of vector: %ld; should be [1:%d]", 
      vector_length, MAX_VECTOR_LENGTH);
    }
    return 0;
  }

  template<class _P>
  inline allinone::Graph * make_boost_adjacency_matrix(allinone::Graph * _g, 
                              long no_vertices, bool copy_data = true){
    LOG(DEBUG, "Allocating adjacency_matrix");
    bool is_directed = _g->get_is_directed();
    LOG(DEBUG, " using %s", is_directed? "directed" : "undirected");
    LOG(DEBUG, " using type: %s", 
      allinone::Graph::get_datatype_s(allinone::Graph::get_datatype<_P>()).c_str());
    switch (_g->get_vector_length()){
      BOOST_PP_REPEAT(MAX_VECTOR_LENGTH, GRAPH_CREATOR, 
        (_g, no_vertices, allinone::Graph::Graph_adjacency_matrix, copy_data))
    default:
LOG(ERROR, "Not supported size of vector: %ld; should be [1:%d]", 
      _g->get_vector_length(), MAX_VECTOR_LENGTH);
    }
    return 0;
  }

#undef BASE 
#undef GRAPH_CREATOR

  //some functions
  int test();
  int test_time(int);

  inline void tmp (){
    allinone::Graph * g = make_boost_adjacency_list<_PFloat>(false, true, 5, 100);
    g = make_boost_adjacency_matrix<char>(false, true, 10, 100);

  }
}

#endif /*GRAH_BOOST_HPP_*/

