#include "graph.hpp"
#include "boost_graphs.hpp"
#include "timer/timer.h"

#include<stdio.h>
#include<unistd.h>

#define NUMBERS_IN_LINE 5

namespace graph{


  using allinone::ArgAtoms;
  using allinone::ArgAtom;
  using allinone::Packages;
  using allinone::Package;
//  using allinone::Graph;
  void save_list(FILE * f, string name, allinone::Graph * g, char delimit){
    for(int i = 0, iend = g->get_no_vertices(); i < iend; i++)
//    for(int j = 0, jend = g->get_is_weighted() ? iend : i + 1; j < jend; j++){
//    To utilize with JointClustering
    for(int j = 0, jend = g->get_is_directed() ? iend : i; j < jend; j++){
      if(g->is_edge(i, j, 0)){
        fprintf(f, "%s%c%s%c%.12g\n", 
//          g->get_name(i), delimit, g->get_name(j), delimit, g->get(i, j, 0));
//          To utilize with JointClustering
//  g->get_name(i), delimit, g->get_name(j), delimit, max(g->get(i, j, 0),0.0));
//  g->get_name(i), delimit, g->get_name(j), delimit, (g->get(i, j, 0)+1)/2.0);
//  g->get_name(i), delimit, g->get_name(j), delimit, fabs(g->get(i, j, 0)));
  g->get_name(i), delimit, g->get_name(j), delimit, 1.0 / g->get(i, j, 0));
      }
    }
  }


  void save_matrix(FILE * f, string name, allinone::Graph * g, char delimit){
    const char * tmp;
    int columns = g->get_vector_length();
    int genes = g->get_no_vertices();
    fprintf(f, "%d%c%d", genes, delimit, columns);
    int count = 0;
    for(int i = 0; i < genes; ++i){
      if(!(count++ % 10)) fprintf(f, "\n%s", g->get_name(i));
      else fprintf(f, "%c%s", delimit, g->get_name(i));
    }
    fprintf(f, "\n\n");

    count=0;
    for(int i = 0; i < genes; ++i){
      tmp = g->get_desc(i);
      if(!(count++ % NUMBERS_IN_LINE)) fprintf(f, "\n%s", tmp?:"");
      else fprintf(f, "%c%s", delimit, tmp?:"");
    }
    fprintf(f, "\n\n");

    count = 0; 
    double line[10] = {0};
    for(int i = 0, iend = genes; i < iend; ++i)
    for(int j = 0; j < iend; ++j)
    for(int k = 0, kend = columns; k < kend; ++k){
      if(count && !(count % NUMBERS_IN_LINE)) {
//        fprintf(f, "\n%.12g%c%.12g%c%.12g%c%.12g%c%.12g%c%.12g%c%.12g%c%.12g%c%.12g%c%.12g", 
//          line[0], delimit, line[1], delimit,
//          line[2], delimit, line[3], delimit,
//          line[4], delimit, line[5], delimit,
//          line[6], delimit, line[7], delimit,
//          line[8], delimit, line[9]
//        );
        fprintf(f, "\n");
        for(int v = 0; v < NUMBERS_IN_LINE; v++) {
          if(v) fprintf(f, "%c", delimit);
          fprintf(f, "%.12g", line[v]);
        }
      }

      line[count++ % NUMBERS_IN_LINE] = g->get(i, j, k);
    }
    fprintf(f, "\n");
    for(int i = 0; i < count % NUMBERS_IN_LINE; i++) {
      if(i) fprintf(f, "%c", delimit);
      fprintf(f, "%.12g", line[i]);
    }
  }

  void save_csv(FILE * f, string name, allinone::Graph * g, char delimit){
    const char * tmp;
    int genes = g->get_no_vertices();

    for(int i = 0; i < genes; ++i)
      fprintf(f, "%c%s", delimit, g->get_name(i));
    fprintf(f, "\n");

    for(int i = 0, iend = genes; i < iend; ++i){
      fprintf(f, "%s", g->get_name(i));
      for(int j = 0; j < iend; ++j)
        fprintf(f, "%c%.12g", delimit, g->get(i, j, 0));
      fprintf(f, "\n");
    }
  }

  void save_mat(FILE * f, string name, allinone::Graph * g, char delimit){
    const char * tmp;
    int genes = g->get_no_vertices();

    for(int i = 0; i < genes; ++i){
      if(i) fprintf(f, " ");
      fprintf(f, "\"%s\"", g->get_name(i));
    }
    fprintf(f, "\n");

    for(int i = 0, iend = genes; i < iend; ++i){
      fprintf(f, "\"%s\"", g->get_name(i));
      for(int j = 0; j < iend; ++j)
        fprintf(f, " %.12g", g->get(i, j, 0));
      fprintf(f, "\n");
    }
  }

  void save_file(string name, allinone::Graph * g, char delimit, string filetype){
    if(!g) return;
    string name2 = allinone::getWriteFilename(name);
    FILE * f = name2 == "-" ? stdout : fopen(name2.c_str(), "w");
    if(!f){
      LOG(WARNING, "Can't open file for write: %s, skipping", name2.c_str());
      return;
    }

    if(filetype == "list")
      save_list(f, name, g, delimit);
    else if(filetype == "csv")
      save_csv(f, name, g, delimit);
    else if(filetype == "r-mat")
      save_mat(f, name, g, delimit);
    else if (filetype == "matrix" || filetype =="") 
      save_matrix(f, name, g, delimit);
    else LOG(ERROR, "Unsupported file type: %s", filetype.c_str());

    if(name2 == "-") fprintf(f, "\n");
    fclose(f);
  }

  allinone::Graph * allocate_graph(bool directed, bool weighted, 
        long vector_length, long no_vertices, string data_type, string type, 
        _PFloat threshold)
  {
    if(data_type == "")  data_type = "pfloat";
    if(type == "") type = "matrix";
#define DATATYPE_IF(r, x, t)                                                  \
    if(data_type == STRINGIFY(t)){                                            \
      if(type == "matrix"){                                                   \
        return new                                                            \
          Graph<t>(directed, weighted, vector_length, no_vertices, threshold);\
      }                                                                       \
      else if (type == "adjacency_list"){                                     \
        return make_boost_adjacency_list<t>(directed,                         \
          weighted, vector_length, no_vertices, threshold);                   \
      }                                                                       \
      else if (type == "adjacency_matrix"){                                   \
        return make_boost_adjacency_matrix<t>(directed,                       \
          weighted, vector_length, no_vertices, threshold);                   \
      }                                                                       \
      else                                                                    \
        LOG(ERROR, "Unsupported graph type %s", type.c_str());                \
    } else
    BOOST_PP_SEQ_FOR_EACH(DATATYPE_IF, ~, DATATYPES) 
      LOG(ERROR, "Unsupported data-type %s", data_type.c_str());
#undef DATATYPE_IF
    return 0;
  }
  
  allinone::Graph * allocate_graph(allinone::Graph * g, long no_vertices, 
      string data_type, string type, bool copy_data)
  {
    if(data_type == "")  data_type = "float";
    if(type == "") type = "matrix";
#define DATATYPE_IF(r, x, t)                                                  \
    if(data_type == STRINGIFY(t)){                                            \
      if(type == "matrix"){                                                   \
        return new                                                            \
          Graph<t>(g, no_vertices);                                           \
      }                                                                       \
      else if (type == "adjacency_list"){                                     \
        return make_boost_adjacency_list<t>(g, no_vertices, copy_data);       \
      }                                                                       \
      else if (type == "adjacency_matrix"){                                   \
        return make_boost_adjacency_matrix<t>(g, no_vertices, copy_data);     \
      }                                                                       \
      else                                                                    \
        LOG(ERROR, "Unsupported graph type %s", type.c_str());                \
    } else

    BOOST_PP_SEQ_FOR_EACH(DATATYPE_IF, ~, DATATYPES) 
#undef DATATYPE_IF
    return 0;
  }

  allinone::Graph * read_list(deque<string> * deq, string name, char delimit,
                           string datatype, string graphtype){
    LOG(ERROR, "SORRY, reading the graph list is not supported yet");
    return 0;
  }
  
  allinone::Graph * read_matrix(deque<string> * deq, string name, char delimit,
                           string datatype, string graphtype){
    const char * tmp;
    size_t lines = deq->size() ;
    if(lines < 7) LOG(ERROR, "File: %s has less than 7 lines what is minimum of the graph file", name.c_str());
    
    const char * tmpstr; 
    long no_vertices = 0, vector_length = 0;
    list<string> * l;
    l = allinone::splitNStrip((*deq)[0], delimit);
    if(l->size() != 2 || !allinone::toInt(l->front(), &no_vertices) 
                      || !allinone::toInt(l->back(), &vector_length))
      LOG(ERROR, "First decoded line: %s, should have 2 integers.", (*deq)[0].c_str());
    delete l;
    if(!(no_vertices > 0 && vector_length > 0) || 
        no_vertices * no_vertices * vector_length > 1000*1000*1000){
      LOG(ERROR, "Wrong number of vertices."
      " The number has to be bigger than 0 and the graph is limited to 1G edges."
      "And is: %ld, %ld", no_vertices, vector_length);
    }
    allinone::Graph * g = allocate_graph(false, true, vector_length, 
                                    no_vertices, datatype, graphtype);
    
    long genes = 0;
    int line = 1; 
    
    allinone::skipEmpty(deq, line, delimit);
    for(l = allinone::splitNStrip((*deq)[line], delimit); 
          line < deq->size() && 
          !(l->size() == 0 || l->size() == 1 && l->front() == "") ; 
          l = allinone::splitNStrip((*deq)[++line], delimit)){
      foreach(string & str, *l){
        if(genes >= no_vertices) genes++; 
        else g->set_names(&(tmp = str.c_str()), 1, genes++);
      } delete l;
      if(line == deq->size() - 1) break;
    } 
    if(line != deq->size() - 1) delete l;
    if(genes > no_vertices) 
      LOG(WARNING, "To much genes names: %ld, expected: %ld", genes, no_vertices);
    if(genes < no_vertices) 
      LOG(ERROR, "To few genes names: %ld, expected: %ld", genes, no_vertices);
    if(deq->size() <= line) 
      LOG(ERROR, "File ended unexpectedly after gene names");

    genes = 0;
    allinone::skipEmpty(deq, line, delimit);
    for(l = allinone::splitNStrip((*deq)[line], delimit); 
          line < deq->size() && 
          !(l->size() == 0 || l->size() == 1 && l->front() == "") ; 
          l = allinone::splitNStrip((*deq)[++line], delimit)){
      foreach(string & str, *l){
        if(genes >= no_vertices) genes++; 
        else g->set_descs(&(tmp = str.c_str()), 1, genes++);
      } delete l;
      if(line == deq->size() - 1) break;
    } 
    if(line != deq->size() - 1) delete l;

    if(genes > no_vertices) 
      LOG(WARNING, "To much genes descriptions: %ld, expected: %ld, skipping...", 
        genes, no_vertices);
    if(deq->size() <= line) 
      LOG(ERROR, "File ended unexpectedly after genes descriptions");

    long edges = 0;
    long no_edges = no_vertices * no_vertices * vector_length;
//    _PFloat * pptr = &g->get(0, 0, 0);
    double val = 0;

    allinone::skipEmpty(deq, line, delimit);
    for(l = allinone::splitNStrip((*deq)[line], delimit); 
          line < deq->size() && 
          !(l->size() == 0 || l->size() == 1 && l->front() == "") ; 
          l = allinone::splitNStrip((*deq)[++line], delimit)){
      foreach(string & str, *l){
        if(edges >= no_edges); 
        else 
          if(allinone::toFloat(str, &val)){
            g->set( edges / vector_length / no_vertices, 
                    edges / vector_length % no_vertices,       
                    edges % vector_length, val);
          }
        else LOG(ERROR, "Can't decode float: %s", str.size());
        ++edges;
      } delete l;
      if(line == deq->size() - 1) break;
    } 
    if(line != deq->size() - 1) delete l;
    if(edges > no_edges)
      LOG(WARNING, "There are %ld edges weights in the graph, expected: %ld ", 
            edges, no_edges);
    if(edges < no_edges)
      LOG(ERROR, "There are %ld edges weights in the graph, expected: %ld ", 
            edges, no_edges);

    return g;
  }

  allinone::Graph * read_file(string name, char delimit, string datatype, 
                                           string graphtype, string filetype){
    allinone::Graph * g = 0;
    deque<string> * deq = allinone::getStringsFromFile(name.c_str());
    if(!deq) LOG(ERROR, "Can't read file: %s", name.c_str());

    if(filetype == "list")
      g = read_list(deq, name, delimit, datatype, graphtype);
    else if (filetype == "matrix" || filetype =="") 
      g = read_matrix(deq, name, delimit, datatype, graphtype);
    else LOG(ERROR, "Unsupported file type: %s", filetype.c_str());
    delete deq;
    return g;
  }

  Graph<_PFloat> * merge(deque<allinone::Graph *> & l){
//    set<const char *, allinone::charptr_less> genes;
//    vector<const char *> exps; exps.reserve(20);
//    foreach(Microarray<_PFloat> * m, l){
//      const char ** strpp = &m->get_name(0);
//      genes.insert(strpp, &strpp[m->get_no_genes()]);
//      strpp = &m->get_exp(0);
//      exps.insert(exps.end(), strpp, &strpp[m->get_vector_length()]);
//    }
//    if(!genes.size() || !exps.size()) return 0;
//
//    Microarray<_PFloat> * mnew = new Microarray<_PFloat>(exps.size(), genes.size());
//    mnew->set_exps(&exps[0], exps.size());
//    mnew->set_names(&vector<const char *>(genes.begin(), genes.end())[0], genes.size());
//
//    foreach(Microarray<_PFloat> * m, l){
//      for(int i = 0, iend = m->get_no_genes(); i < iend; ++i){
//        const char * name = m->get_name(i);
//        for(int j = 0, jend = m->get_vector_length(); j < jend; ++j){
//          mnew->get(name, m->get_exp(j)) = m->get(i, j);
//        }
//      }
//    }
//
//    return mnew;
    return 0;
  }


  InterfaceFunction{

    if(package->interface && package->interface->arg != InterfaceFlow::None)
      return 0;
    int n = 1;
    char str[100];
    sprintf(str, "%s:p[%d]", package->name.c_str(), package->packageNumber);
    timer::Timer t(str);
    t.start();
      ArgAtoms outputs;
      deque<allinone::Graph *> graphs;
      Package * p;
      Interface * interface;

      char delimit = ';';
      string datatype  = "";
      string graphtype = "";
      string filetype = "";
      

      for(ArgAtoms::iterator it = package->argNode.begin(); 
            it != package->argNode.end(); ++it){
        if((*it)->argAtomType & (ArgAtom::LongSwitch | ArgAtom::ShortSwitch)){
          if((*it)->value == "-i" || (*it)->value == "--input"){
            *++it;
            switch((*it)->argAtomType){
            case ArgAtom::ToProcess:
            case ArgAtom::PackageNumber:
              getInterfaces(&graphs, (*it), InterfaceFlow::Graph);
              break;
            case ArgAtom::Arg:
              graphs.push_back(read_file((*it)->value, delimit, datatype, graphtype, filetype));
            break;
            }
          }else if((*it)->value == "-o" || (*it)->value == "--output"){
            outputs.push_back(*++it);
          }else if((*it)->value == "-g" || (*it)->value == "--graph-type"){
            graphtype =(*++it)->value;
          }else if((*it)->value == "-t" || (*it)->value == "--data-type"){
            datatype = (*++it)->value;
          }else if((*it)->value == "-f" || (*it)->value == "--file-type"){
            filetype = (*++it)->value;
          }else if((*it)->value == "-n" || (*it)->value == "--number-of"){
            n = atoi((*++it)->value.c_str());
            if(!(n >= 1)){
              LOG(ERROR, "Wrong argument of \"--number-of\" switch: %d", 
                        (*it)->value.c_str());
            }
          }else if((*it)->value == "-d"){
            delimit = allinone::parseDelimiter((*++it)->value);
          }
        }
      }

      allinone::Graph * graph;
      list<Interface *> interfaces;
      switch(graphs.size()){
      case 0:
        LOG(ERROR, "No input source specified for package[%d]: %s", 
          package->packageNumber, package->name.c_str());
      case 1:
        graph = graphs.front();
        if((graphtype == "" || graphtype == graph->get_graphtype_s()) && 
           (datatype == "" || datatype == graph->get_datatype_s()));
        else 
          graph = allocate_graph(graph, graph->get_no_vertices(), datatype, graphtype);
         break;
      default:
        graph = merge(graphs);
        foreach(allinone::Graph * g, graphs) 
          interfaces.push_back((Interface *) g);
        allinone::freeFreeable(interfaces);
      }

    graph->readOnly = true;
    graph->freeAble = false;

    package->interface = graph;

    foreach(ArgAtom * o, outputs){
      save_file(o->value, graph, delimit, filetype);
//      printf("Values of graph (%s, %s, %s, %s, %s):\n\t%g, %g, %g, %g, %g\n", 
//        graph->get_name(0), graph->get_name(1), graph->get_name(2), 
//        graph->get_name(3), graph->get_name(4), 
//        graph->get(0,0,0), graph->get(0,1,0), graph->get(0,2,0), 
//        graph->get(0,3,0), graph->get(0,4,0));
    }
    
    t.stop();
    return 0;
  }
}



