#ifndef GRAPH_INTERFACE_H_
#define GRAPH_INTERFACE_H_

#include<boost/preprocessor.hpp>
#include<interface.h>

  //SHOULD BE ALREADY INCLUDED
  //#include <interface.h>

//  #define DATATYPES (pfloat)(char)(uchar)(short)(ushort)(int)(uint)(long)(ulong)
  #define DATATYPES (pfloat)(char)(short)(int)(long)
  #define DATATYPES_TOSTR(r, x, i, t) x STRINGIFY(t)

  typedef unsigned char uchar;
  typedef unsigned int uint;
  typedef unsigned short ushort;
  typedef unsigned long ulong;
  typedef _PFloat pfloat ;

  const ArgsDef  args_graph[] = {
      {'t', "data-type", arg_req_arg, "data type of the graph weights;"
        " Possible values are:\n\t\t" 
        STRINGIFY(BOOST_PP_SEQ_HEAD(DATATYPES))
        BOOST_PP_SEQ_FOR_EACH_I(DATATYPES_TOSTR, ", ", BOOST_PP_SEQ_TAIL(DATATYPES)) 
"\n\t\tThe default is [pfloat] and depending, on the compilation it is float or double.",
        STRINGIFY(BOOST_PP_SEQ_HEAD(DATATYPES))
        BOOST_PP_SEQ_FOR_EACH_I(DATATYPES_TOSTR, "|", BOOST_PP_SEQ_TAIL(DATATYPES)) 
      },
      {'g', "graph-type", arg_req_arg, "graph representation type; Possible values are:"
"\n\t\tmatrix (self implemented), adjacency_list, adjacency_matrix (boost)" 
"\n\t\tThe default is [matrix]. In general, matrix-based graphs are better"
"\n\t\tfor dense graphs and adjacency_list is better for sparse graphs.",
      "matrix|adjacency_list|adjacency_matrix"
      },
      {'f', "file-type", arg_req_arg, 
      "the way the data should be read/stored in the file; "
      "correct values are: 'matrix' - the whole matrix of possible "
      " connections', best for dense graphs; 'list' - delimited edges "
      "with corresponding weights, 'csv' - a big matrix, "
      "'r-mat' - similar to 'csv', but following the R-cran format;",
      "matrix|list|csv|r-mat"},
      {'d', "", arg_req_arg, 
"CSV input/output file delimitier; works on files read after this switch; default: ';'" },
      {0, "", arg_no_arg, "A package that reads and writes graphs", ""}};


  namespace graph{
    InterfaceFunction;
  
    const InterfaceFlow interfaceFlow(
      InterfaceFlow::InterfaceArg(InterfaceFlow::Graph | InterfaceFlow::Multiple),
      InterfaceFlow::Graph);
  }


#endif /*GRAPH_INTERFACE_H_*/
