#include "graph_matrix.hpp"
#include "timer/timer.h"

#include<stdio.h>
#include<unistd.h>

namespace graph_matrix{

  using allinone::Graph;

  int test()
  {
    allinone::Graph<_PFloat> * g = new Graph<_PFloat> (true, true, 2, 8);
  
    const char * names[] = {"jas", "malgosia", "gosia", "mosia", "srosia"};
    g->set_names(names, 5, 2);
    
    printf("verts, from begining: \n\t");
    for(Graph<_PFloat>::iter vi = g->begin(); vi!= g->end(); vi++){
      printf("'%s': ", (*vi).get_name());
    }
  
    printf("\n\nverts, from end: \n\t");
    for(Graph<_PFloat>::riter vi = g->rbegin(); vi!= g->rend(); vi++){
      printf("'%s': ", (*vi).get_name());
    }
    
    for(int j = 0, jend = g->get_no_vertices(); j < jend; j++)
    for(int i = 0, iend = g->get_no_vertices(); i < iend; i++){
      (*g)[j][i][0] = i + 100 * j;
    }
  
    printf("\n\nMatrix:");
    int i = 0, iend = g->get_no_vertices();
  
    for(Graph<_PFloat>::eiter<allinone::None> ei = g->ebegin<allinone::None>(); 
        ei != g->eend<allinone::None>(); ++ei){
      if(!(i++ % iend))
        printf("\n\t");
      printf("(%d,%d)=%.0f ", ei->from, ei->to, (*ei)[0]);
    }

    printf("\n\nMatrix (Reverse):");
    for(Graph<_PFloat>::eiter<allinone::Reverse> ei = g->ebegin<allinone::Reverse>(); 
        ei != g->eend<allinone::Reverse>(); ++ei){
      if(!(i++ % iend))
        printf("\n\t");
      printf("(%d,%d)=%.0f ", ei->from, ei->to, (*ei)[0]);
    }

    printf("\n\nMatrix (Succs[2]):");
    for(Graph<_PFloat>::eiter<allinone::Successors> ei = g->ebegin<allinone::Successors>(2); 
        ei != g->eend<allinone::Successors>(2); ++ei){
      if(!(i++ % iend))
        printf("\n\t");
      printf("(%d,%d)=%.0f ", ei->from, ei->to, (*ei)[0]);
    }
    printf("\n");

    printf("\n\nMatrix (Succs rev[2]):");
    for(Graph<_PFloat>::eiter<allinone::Edge_iterator_Type(allinone::Successors | allinone::Reverse)> ei = 
        g->ebegin<allinone::Edge_iterator_Type(allinone::Successors | allinone::Reverse)>(2); 
        ei != g->eend<allinone::Edge_iterator_Type(allinone::Successors | allinone::Reverse)>(2); ++ei){
      if(!(i++ % iend))
        printf("\n\t");
      printf("(%d,%d)=%.0f ", ei->from, ei->to, (*ei)[0]);
    }
    printf("\n\nMatrix (Pred[2]):");
    for(Graph<_PFloat>::eiter<allinone::Predecessors> ei = g->ebegin<allinone::Predecessors>(2); 
        ei != g->eend<allinone::Predecessors>(2); ++ei){
      if(!(i++ % iend))
        printf("\n\t");
      printf("(%d,%d)=%.0f ", ei->from, ei->to, (*ei)[0]);
    }
    printf("\n");

    printf("\n\nMatrix (Pred rev[2]):");
    for(Graph<_PFloat>::eiter<allinone::Edge_iterator_Type(allinone::Predecessors | allinone::Reverse)> ei = 
        g->ebegin<allinone::Edge_iterator_Type(allinone::Predecessors | allinone::Reverse)>(2); 
        ei != g->eend<allinone::Edge_iterator_Type(allinone::Predecessors | allinone::Reverse)>(2); ++ei){
      if(!(i++ % iend))
        printf("\n\t");
      printf("(%d,%d)=%.0f ", ei->from, ei->to, (*ei)[0]);
    }
    printf("\n");
   
    return 0;
  }

  int test_time(int len)
  {
    allinone::Graph<_PFloat> * g = new Graph<_PFloat> (true, true, 2, len);
      
    timer::Timer t("graph_timing");t.start();
    for(int j = 0; j < 10; j++){
      int i=0;
      for(Graph<_PFloat>::eiter<allinone::None> ei = g->ebegin<allinone::None>(), 
          eiend = g->eend<allinone::None>(); ei!= eiend; ++ei){
        (*ei)[0] = i++;
      }
      for(Graph<_PFloat>::eiter<allinone::None> ei = g->ebegin<allinone::None>(), 
          eiend = g->eend<allinone::None>(); ei!= eiend; ++ei){
        (*ei)[0] += 2.0;
      }
    }
    t.stop();

    timer::Timer t2("graph_timing2");t2.start();
    for(int j = 0; j < 10; j++){
      int i=0;
      for(int vj = 0; vj < len; vj++)
      for(int vi = 0; vi < len; vi++){
        (*g)[vj][vi][0] = i++;
      }
      for(int vj = 0; vj < len; vj++)
      for(int vi = 0; vi < len; vi++){
        (*g)[vj][vi][0] += 2.0;
      }
    }
    t2.stop();

    delete g;

    _PFloat * gp = new _PFloat[len * len * 2];
    timer::Timer t3("graph_timing_arr"); t3.start();
    for(int j = 0; j < 10; j++){
      for(int i=0, iend = len*len * 2; i < iend; i+=2)
        gp[i] = i;
      for(int i=0, iend = len*len * 2; i < iend; i+=2)
        gp[i] += 2;
    }
    t3.stop();
           
    timer::Timer t4("graph_timing_ptr"); t4.start();
    for(int j = 0; j < 10; j++){
      int i = 0;
      for(_PFloat * p = gp, * pend = &gp[len * len * 2]; p != pend; p+=2)
        *p=i++;
      for(_PFloat * p = gp, * pend = &gp[len * len * 2]; p != pend; p+=2)
        *p+=2;
    }
    t4.stop();

    printf("Timings:\n\tgraph it: %f\n\tgraph arr: %f\n\tarray: %f\n\tptr: %f\n",
              timer::TimerSet::getInstance()->getSummary("graph_timing"),
              timer::TimerSet::getInstance()->getSummary("graph_timing2"),
              timer::TimerSet::getInstance()->getSummary("graph_timing_arr"),
              timer::TimerSet::getInstance()->getSummary("graph_timing_ptr"));
    
    return 0;
  }

}



