#include "config.h"
#include <interface.h>
#ifdef HAVE_TIMER_INTERFACE_H
#  include"timer/timer.h"  
#endif

#include <sys/stat.h>
#include <sys/types.h>
#include<list>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<string>
#include<sstream>

using namespace std;

Verbosity verbosity = DEBUG;

namespace allinone {

  static const char ** strings = 0;
  static int curstring = 0;
  static int curchar = 0;
  static int isfromargs = 0;
  ChangeCase changeCase = no;


  ArgAtoms argsAtomised;
  Packages packages;
  PackagesByNumbers packagesByNumbers;
  string outDir = "./";

  void help_package(const char * package, const char * print = 0){
    if(print) fprintf(stderr, "%s\n", print); 
    else fprintf(stderr, "Package: %s\n", package); 

    const ArgsDef * startptr = packageArgs.find(package)->second;
    const ArgsDef * adptr;
    for (adptr = startptr; !(adptr->shortSwitch == 0 && 
                 adptr->longSwitch == ""); ++adptr);
    if(adptr->help != "")
      fprintf(stderr, "%s\n", adptr->help.c_str()); 
      
    fprintf(stderr, "Options: \n"); 
    for (adptr = startptr; !(adptr->shortSwitch == 0 && 
                 adptr->longSwitch == ""); ++adptr){
      if(adptr->shortSwitch != 0 && adptr->longSwitch != ""){
        fprintf(stderr, "\t-%c, --%s %s\n\t\t%s\n", adptr->shortSwitch, 
          adptr->longSwitch.c_str(), 
          adptr->arg & arg_req_arg ? "<arg>" : 
            adptr->arg & arg_req_arg ? "[<arg>]" : "", adptr->help.c_str()); 
      }else if(adptr->shortSwitch != 0){
        fprintf(stderr, "\t-%c %s\n\t\t%s\n", adptr->shortSwitch, 
          adptr->arg & arg_req_arg ? "<arg>" : 
            adptr->arg & arg_req_arg ? "[<arg>]" : "", adptr->help.c_str()); 
      }else if(adptr->longSwitch != ""){
        fprintf(stderr, "\t--%s %s\n\t\t%s\n", adptr->longSwitch.c_str(), 
          adptr->arg & arg_req_arg ? "<arg>" : 
            adptr->arg & arg_req_arg ? "[<arg>]" : "", adptr->help.c_str()); 
      }
    }
  }

  void help(){
    
    fprintf(stderr, 
"This is program CLAIM (CLuster Analysis Integration Method) ver: %s. \n"
"If you need more detailed doc. or want to report a bug, please contact:\n"
"Marek Blazewicz <mblazewicz@cs.put.poznan.pl>. \n"
"This program is distributed under GNU GPL.\n"
"The main purpose of this work is to create an easy tool for biologists to\n"
"manipulate different types of biological data and help identifying functional\n"
"modules, i.e. sets of genes performing similar tasks in living organisms. \n\n"
"It is easily extensible and configurable. User may add his own packages to\n"
"process the data. The only requirement from user is create a package (directory\n"
"with sources in src/ directory and implement a function <package>::process.\n"
"This function needs to support \"common\" interface, parse input arguments and\n"
"return some output. More detailed documentation will come...\n\n"
"User's task is to define a data flow. By data flow I mean a graph-like \n"
"dependencies passing the output of one package as an input of another. The user\n"
"might define an input of a package in three different ways: \n"
"<filename>, {-p <package> args}, \\<package number>.\n"
"In general the program launch looks like this:\n"
"./allinone <program options> -p <package name> <package options> \n"
"  [-p <package name> -i{-p <package name> <options>: <options2>}-p <name> -i \\1 \n"
"  where ':' is depicting that the program should use the same package as the \n"
"  former but with different arguments: <options2>.\n\n"
"The arguments can be passed to the program as common arguments or, for larger\n"
"data flows, passed by stdin. \n"
"It must be stated at this point that the current version of the program doesn't\n"
"fully support the conversion of arguments by the interpreter. For instance:\n"
"if you pass a 'string' as an argument it is recognized by the program as a\n"
"string only if it has a quotation mark, whereas interpreter suppresses it and\n"
"passes it as characters inside the string but as one argument to the program \n"
"(including whitespace). In current version the program doesn't support such \n"
"strings and user needs to be aware of it. It will be corrected in the near future.\n\n"
"Below one can find the full list of available packages and its options.\n\n", 
      PACKAGE_VERSION);
  
    help_package("allinone", "Program options:");
    help_package("common", "Common options for all packages:");
    for (PackageArgs::const_iterator it = packageArgs.begin(); 
          it != packageArgs.end(); ++it){
      if(it->first != "allinone" && it->first!="common"){
        help_package(it->first.c_str());
      }
    }
    exit(0);
  }

  int process(allinone::Package * package){

    Verbosity oldverbosity = verbosity; 

    for(ArgAtoms::iterator it = package->argNode.begin(); 
            it != package->argNode.end(); ++it){
      if((*it)->value == "-v" || (*it)->value == "--verbose"){
        if((*++it)->argAtomType == ArgAtom::Arg){
          if((*it)->value == "debug") verbosity = DEBUG;
          else if((*it)->value == "info") verbosity = INFO;
          else if((*it)->value == "notice") verbosity = NOTICE;
          else if((*it)->value == "warning") verbosity = WARNING;
          else if((*it)->value == "error") verbosity = ERROR;
          else LOG(ERROR,"Unsupported verbosity value: %s", (*it)->value.c_str());
        }else{
          --it;
          verbosity = INFO;
        }
      }else if((*it)->value == "-h" || (*it)->value == "--help"){
        help();
      }else if((*it)->value == "-O" || (*it)->value == "--outdir"){
        if((*++it)->argAtomType != ArgAtom::Arg)
          LOG(ERROR,"Outdir requires argument.");
        outDir = (*it)->value + '/';
        if(mkdir(outDir.c_str(), 0755))
          LOG(WARNING,"Cannot create dir: %s, hopefully it is dir already...", outDir.c_str());
      }
    }

    if(verbosity >= INFO && oldverbosity < INFO){
      for(Packages::iterator it = packages.begin(); it != packages.end(); it++){
        (*it)->checkConsistency(0);
      }
    }
    return 0;
  }

  static int getChar(){
    int curr = EOF;
    if(strings[curstring]){
      curr = strings[curstring][curchar];
      do{
        curchar++; 
        if(!strings[curstring][curchar]){
          curchar = 0;
          curstring++;
        }
      }while(strings[curstring] && !strings[curstring][curchar]);
    }
    return curr;
  }
  
  static int ungetChar(int n = 1){
    int curr = 0;
    if(curstring || curchar){
      do{
        curchar--; 
        if(curchar < 0){
          curstring--;
          curchar = strlen(strings[curstring]) - 1;
        }
        if(strings[curstring][curchar]) { --n; ++curr; }
      }while(curstring && curchar && n);
    }
    return curr;
  }

  
  ArgAtom getNext(){
    ostringstream os;
    char isstring = 0, isbackslash = 0;
    int written = 0;
    int curchar, tmpchar, prevchar = 0; 
    int running = 1;
    static int opened = 0;
    string str;
    ArgAtom aa;
    static list<ArgAtom> aapending;
    if(aapending.size()){
      aa=aapending.front();
      aapending.pop_front();
      goto end;
    }
    while(running && (curchar = getChar()) != EOF){
      switch(curchar){
        case ' ':
        case '\n':
        case '\r':
        case '\t':
          if(isbackslash){
            os.put(' '); written++;
            if(curchar == '\n' || curchar == '\r'){
              tmpchar = getChar();
              if(curchar != tmpchar && (tmpchar == '\n' || tmpchar == '\r'));
              else ungetChar();
            }
          }
          else if(isstring){ os.put(curchar); written++; }
          else if(written)
            running = false;
        break;

        case '\'':
        case '"':
          if(isbackslash){
            if(isstring == curchar){ os.put(curchar); written++; }
            else if(isstring) { os.put('\\'); os.put(curchar); written+=2; }
            else { os.put(curchar); written++; }
          }else if(isstring == curchar)
            isstring = false;
          else if(isstring){
            os.put(curchar); written++;
          }else isstring = curchar; 
        break;

        case 'n':
        case 'r':
        case 't':
          if(isstring == '"' && isbackslash){
            os << string("\\") + (char)curchar; written++;
          }else{
            os.put(curchar); written++;
          }
        break;

        case '{':
        case '}':
        case ':':
          if(isstring){
            if(isbackslash) { os.put('\\'); written++; }
            os.put(curchar); written++;
          }else if (isbackslash){
            os.put(curchar); written++;
          }else{
            if(written){ ungetChar();  }
            else{
              os.put(curchar); written++;
              aa.argAtomType = ArgAtom::Operator;
              switch(curchar){
                case '{': opened++; break;
                case '}': opened--; break;
              }
            }
            running=false;
          }
        break;

        case '\\':
          if(isbackslash){ os.put(curchar); written++; }
          isbackslash^='\\';
        break;

        case '0': case '1': case '2': case '3': case '4': 
        case '5': case '6': case '7': case '8': case '9': 
          if(!written && isbackslash) 
            aa.argAtomType = ArgAtom::PackageNumber;
          os.put(curchar); written++;
        break;

        case '-':
          if(!written && !isbackslash) 
            aa.argAtomType = ArgAtom::ShortSwitch;
          else if(written == 1 && !isbackslash 
                    && aa.argAtomType == ArgAtom::ShortSwitch) 
            aa.argAtomType = ArgAtom::LongSwitch;
          os.put(curchar); written++;
        break;

        default:
          if(isstring && isbackslash){
            os.put('\\'); os.put(curchar); written++;
          }else {
            os.put(curchar); written++;
          }
      }
      if(isbackslash && curchar!='\\')
        isbackslash = 0;
    }
    os.flush();
    if(isbackslash) LOG(ERROR, "Unterminated backslash in parameters");
    if(isstring) LOG(ERROR, "Unterminated string: %c in parameters",isstring);
    
    if(opened < 0)
      LOG(ERROR, "There are %d closed, but not opened parentheses in the argument list",-opened);

    str = os.str();
    if(aa.argAtomType == ArgAtom::None){
      if(str.size() > 0)
        aa.argAtomType = ArgAtom::Arg;
      else if(opened > 0)
        LOG(ERROR, "There are %d opened, but not closed parentheses in the argument list", opened);
    }
    else if(aa.argAtomType == ArgAtom::PackageNumber){
      if(str.find_first_not_of("0123456789") != string::npos){
        aa.argAtomType = ArgAtom::Arg;
        str=string("\\") + str;
      }
    }else if(aa.argAtomType == ArgAtom::ShortSwitch){
      if(str.size() == 1)
        aa.argAtomType = ArgAtom::Arg;
      else if(str.size() > 2){
        for(int i = 2; i < str.size(); i++)
          aapending.push_back(ArgAtom(ArgAtom::ShortSwitch, string("-") + str[i]));
        str = str.substr(0, 2);
      }
    }
    aa.value = str;
    end: 
#ifdef __DEBUG__
    LOG(DEBUG, "found atom %s:'%s'", ArgAtomType2Str.find(aa.argAtomType)->second, aa.value.c_str());
#endif
    return aa;
  }

   
  ArgAtoms::iterator findArgsEnd (ArgAtoms::iterator beg, ArgAtoms::iterator end){
    int opened = 0;
    if(beg != end)
      for(ArgAtoms::iterator it = ++beg; it!=end; ++it){
        if((*it)->argAtomType == ArgAtom::Operator){
          if((*it)->value == "}"){
            if(--opened < 0) return it;
          }
          else if((*it)->value == "{")
            ++opened;
          else if((*it)->value == ":" && !opened)
            return it;
        }
        else if(!opened && (*it)->isPackageSpecifier())
          return it;
      }
    return end;
  }
  
  Package::Package(ArgAtoms::iterator beg, ArgAtoms::iterator end, 
                            int & packageNumber, const string & name) : interface(0){
    ArgAtoms::iterator it = beg, tmpit, farend = end;
    int opened = 0, next = 0;
    ArgAtom * aa = 0;

    this->packageNumber = packageNumber++; 
    if(!this->packageNumber){
      if((*it)->isPackageSpecifier()) end = it;
      else end = findArgsEnd(it, end);
      this->name = string("allinone");
    }else {
      if((*it)->isPackageSpecifier()){
        if(++it != end && (*it)->argAtomType == ArgAtom::Arg)
          this->name = (*it++)->value;
        else{
          LOG(ERROR, "Each arg sequence must start from package specification.\n"
                   "\t-p/--package <package_name>");
        }
      }else if (name != "")
        this->name = name;
    }
      
    if(packageArgs.find(this->name) == packageArgs.end() || this->name=="")
      LOG(ERROR, "Cannot find package: %s", this->name.c_str());
    
    for(; it!=end; it++){
      switch ((*it)->argAtomType){
        case ArgAtom::Operator: 
          if((*it)->value == "{") opened++;
          else if((*it)->value == "}") {
            if(!--opened){
              if(!aa || aa->argAtomType != ArgAtom::ToProcess){
                LOG(ERROR, "Cannot add empty specification of the input");
              }
              argNode.push_back(aa);
              aa = 0;
            }
          }
          else if((*it)->value == ":"){ 
            if(opened && aa->packages.size());// name = aa.packages.back().name;
            else LOG(ERROR, "Can't specify : parameter at the first package in sequence");
          }
        break;

        case ArgAtom::LongSwitch:
        case ArgAtom::ShortSwitch:
          if(opened){
            if(!aa || aa->argAtomType == ArgAtom::None) aa=new ArgAtom(ArgAtom::ToProcess);
            if(aa->argAtomType != ArgAtom::ToProcess) 
              LOG(ERROR, "Error while parsing the arguments");

            tmpit = findArgsEnd(it, end);
            aa->packages.push_back(new Package(it, tmpit, packageNumber, 
                aa->packages.size() ? aa->packages.back()->name : ""));
            --(it = tmpit);
            break;
          }else if((*it)->isPackageSpecifier()){
            LOG(ERROR, "Some error occurred; package specifier found in wrong place");
          }
        
        default:
          if(opened)
            LOG(ERROR, "Some error occured; Argument: %s and "
                       "each package should start from switch", (*it)->value.c_str());

          argNode.push_back(*it);
      }
    }
    
    if(opened)
      LOG(ERROR, "Unclosed parenthesis, cannot proceed...");

    if(!this->packageNumber){
      for(it = end; it != farend; it = tmpit){
        string tmpname = "";
        if((*it)->argAtomType == ArgAtom::Operator){
          if((*it)->value == ":") tmpname = packages.back()->name;
          else LOG(ERROR, "Wrong operator found: %s", (*it)->value.c_str());
          it++;
        }
        tmpit = findArgsEnd(it, farend);
        packages.push_back(new Package(it, tmpit, packageNumber, tmpname));
      }
    }
  }

  void Package::checkConsistency(int level){
    char * backspace;
    ArgArg argArg= arg_no_arg;
    const ArgsDef * adptr;
    const ArgsDef * packAdptr = packageArgs.find(this->name)->second;
    const ArgsDef * commAdptr = packageArgs.find("common")->second;
    int num = 0;
    packagesByNumbers[packageNumber] = this;
    if(!packageInterface.find(name)->second){
      LOG(ERROR, "The package %s exists but no function to process data found in it.", name.c_str());
    }
    backspace = new char[level * 2 + 2];
    backspace[0] = ' ';
    for(int i = 0; i < level; i++){ 
      backspace[i*2+1] = '|';
      backspace[i*2+2] = ' ';
    }
    backspace[level * 2 + 1] = 0;
    backspace[level * 2    ] = ' ';

    if(level)
      LOG(INFO, "%s|-Package[%d]=%s with args:", backspace, this->packageNumber, name.c_str());
    else 
      LOG(INFO, "%sPackage[%d]=%s with args:", backspace, this->packageNumber, name.c_str());

    for(ArgAtoms::iterator it = argNode.begin(); it != argNode.end(); it++){
      switch((*it)->argAtomType){
      case ArgAtom::ShortSwitch: 
        for(adptr = packAdptr; adptr->shortSwitch != 0 && adptr->longSwitch != "" 
            && adptr->shortSwitch != (*it)->value[1]; ++adptr);
        if(adptr->shortSwitch == 0)
        for(adptr = commAdptr; adptr->shortSwitch != 0 && adptr->longSwitch != "" 
            && adptr->shortSwitch != (*it)->value[1]; ++adptr);
        if(adptr->shortSwitch == 0)
          LOG(ERROR, "There is no switch %s in package %s", (*it)->value.c_str(), name.c_str());
        argArg = adptr->arg;

        LOG(INFO, "%s|-Switch : %s", backspace, (*it)->value.c_str()); 
        continue;

      case ArgAtom::LongSwitch:
        for(adptr = packAdptr; adptr->shortSwitch != 0 && adptr->longSwitch != "" 
            && adptr->longSwitch != (*it)->value.substr(2); ++adptr);
        if(adptr->longSwitch == "")
        for(adptr = commAdptr; adptr->shortSwitch != 0 && adptr->longSwitch != "" 
            && adptr->longSwitch != (*it)->value.substr(2); ++adptr);
        if(adptr->longSwitch == "")
          LOG(ERROR, "There is no switch %s in package %s", (*it)->value.c_str(), name.c_str());
        argArg = adptr->arg;

        LOG(INFO, "%s|-Switch : %s", backspace, (*it)->value.c_str()); 
        continue;

      case ArgAtom::Arg:
        if(!(argArg & (arg_opt_arg | arg_req_arg)))
          LOG(ERROR, "Arg '%s' of package '%s' has no switch it is passed to", 
                (*it)->value.c_str(), name.c_str());
        argArg = arg_no_arg;
        LOG(INFO, "%s|-Arg    : %s", backspace, (*it)->value.c_str()); break;

      case ArgAtom::PackageNumber:
        if(!(argArg & (arg_opt_arg | arg_req_arg)))
          LOG(ERROR, "Package number %s has no switch it is passed to", (*it)->value.c_str());
        argArg = arg_no_arg;
        num = strtol((*it)->value.c_str(), 0, 0);
        if(!(num > 0 && num < packagesByNumbers.size())){
          LOG(ERROR, "Wrong package number: %d; the package number should be between 1 and %d",
            num, packagesByNumbers.size());          
        }
        LOG(INFO, "%s|-Package: %d", backspace, num); break;

      case ArgAtom::ToProcess:
        if(!(argArg & (arg_opt_arg | arg_req_arg)))
          LOG(ERROR, "Processed package has no switch it is passed to");
        argArg = arg_no_arg;
        for(Packages::iterator it2 = (*it)->packages.begin(); 
            it2 != (*it)->packages.end(); it2++){
          (*it2)->checkConsistency(level + 1);
        }
      break;

      default:
        LOG(ERROR, "%s Wrong arg type: %s", backspace, 
          ArgAtomType2Str.find((*it)->argAtomType)->second); 
      }

      if(argArg & arg_req_arg)
        LOG(ERROR, "Switch %s of package %s requires argument, and it was not provided.", 
          (*(--it)++)->value.c_str(), name.c_str()); 
      argArg = arg_no_arg;
    }
    if(argArg & arg_req_arg)
      LOG(ERROR, "Switch %s of package %s requires argument, and it was not provided.", 
        argNode.back()->value.c_str(), name.c_str());

    delete [] backspace;
  }

  void atomizeArguments(const int argc, const char ** argv){
       
    timer::Timer t("parsing arguments");
    t.start();
    int i = 0;

    if(argc > 1){
      strings = (const char **) malloc(sizeof(char*) * (argc - 1) * 2);
      for(i = 0; i < argc - 1; i++){
        strings[i * 2] = argv[i + 1];
        strings[i * 2 + 1] = (char*) malloc(sizeof(char) * 2);
        strcpy((char *)strings[i * 2 + 1], "\n");
      }
      i*=2;
      isfromargs=1;
    }
    else{
      list<char *> tmp;
      char * str = (char*)malloc(1000);
      int n = 0;
      while(n = fread(str, 1, 999, stdin)){
        str[n] = 0;
        tmp.push_back(str);
        str = (char*)malloc(1000);
      }
      free(str);
      strings = (const char **) malloc(sizeof(char*) * (tmp.size() + 1));
      for(list<char*>::iterator it = tmp.begin(); it!=tmp.end() ; it++, i++)
        strings[i] = *it;
    }
    strings[i] = 0;

    ArgAtom aa;
    while((aa = getNext()).argAtomType != ArgAtom::None){
      argsAtomised.push_back(new ArgAtom(aa));
    }

    int packageNumber = 0;
    Package * p = new Package(argsAtomised.begin(), argsAtomised.end(), packageNumber);
    packages.push_front(p);
    if(packageNumber){
      packagesByNumbers.resize(packageNumber);
    }

    for(Packages::iterator it = packages.begin(); it != packages.end(); it++){
      (*it)->checkConsistency(0);
    }
    t.stop();
  }
}
