#include"timer/timer.h"
#include"graph_interface.hpp"
#include"results.hpp"
#include"multiple.hpp"
#include"constants.h"
#include<stdio.h>
#include<stdlib.h>
#include<assert.h>
#include<set>
#include<limits>
#include<sys/time.h>
#include<vector>
#include<string>
#include<algorithm>
#include<boost/lexical_cast.hpp>

using allinone::ArgAtoms;
using allinone::ArgAtom;
using allinone::Packages;
using allinone::Package;
using allinone::toInt;
using allinone::toString;
using allinone::Graph;
using allinone::Results;
using allinone::Multiple;

namespace kmeans_euc{
  Results * kmeans_lloyd(Graph * g, int n_of_clusters, int iterations, string);
  Results * kmeans_r(Graph *, int, int, int, string);
}

namespace kmeans{
  
  typedef std::vector<int> assigned_t;
  typedef std::set<int> cluster_t;
  typedef std::vector<cluster_t> clusters_t;
  typedef std::vector<double> scores_t;

  static int seed_initialized = 0;

  void kmeans_store_assignment(assigned_t * assigned, string name){
    name = allinone::getWriteFilename(name);
    FILE * f = fopen(name.c_str(), "w");
    if(!f) LOG(ERROR, "Can't open file: %s", name.c_str());
    for(int i = 0, iend = assigned->size(); i<iend; ++i){
      if(!(i % 5)){
        if(i!=0) fprintf(f,"\n");
        fprintf(f,"%d", assigned->at(i) + 1);
      }else fprintf(f," %d", assigned->at(i) + 1);
    }
    fclose(f);
  }

  void kmeans_calculate_distance(Graph * g, cluster_t * cluster, int from, double * distance){
    double ret = 0;
    foreach(int to, *cluster) {
      if(to != from)
        ret += std::abs<double>(g->get(from, to, 0));
    }
    *distance = ret / cluster->size();
//    *distance = ret / (cluster->size() + 
//        (cluster->find(from) == cluster->end() ? 0 : 1));
  }

  void kmeans_recalculate_sumdistance(Graph * g, cluster_t * cluster, clusters_t * clusters, double * score){
    *score = 0;
    foreach(cluster_t & cl, *clusters){
      if(&cl != cluster)
      foreach(int from, *cluster){
        double distance = 0;
        kmeans_calculate_distance(g, &cl, from, &distance);
        *score += distance;
      }
    }
    *score /= cluster->size() * (clusters->size() - 1);
  }

  void kmeans_recalculate_sumdistances(Graph * g, clusters_t * clusters, scores_t * scores){
    for(int i = 0, iend = clusters->size(); i < iend; ++i)
      kmeans_recalculate_sumdistance(g, &clusters->at(i), clusters, &scores->at(i));
  }

  void kmeans_assign_seeds(Graph * g, clusters_t * clusters, assigned_t * assigned, string in_assigned){
    int n_of_clusters = clusters->size();
    int nverts = g->get_no_vertices();
    if(in_assigned == ""){
      for(int i = 0, v; i < n_of_clusters; ++i)
      {
        do{ v = rand() % nverts; }while(assigned->at(v) != -1);
        clusters->at(i).insert(v); 
        assigned->at(v) = i;
      }
    }
    else{
      deque<string> * deq = allinone::getStringsFromFile(in_assigned.c_str());
      if(!deq) LOG(ERROR, "Can't read file: %s", in_assigned.c_str());
      long gene_no = 0, cluster = 0;
      for(int i = 0; i < deq->size(); ++i){
        list<string> * l;
        l = allinone::splitNStrip(deq->at(i), ' ');
        foreach(string & str, *l){
          if(gene_no >= nverts) 
  LOG(ERROR, "File %s should contain %d number of intergers, which is number of clustered genes",
                       in_assigned.c_str(), nverts);
          if(!toInt(str, &cluster) || cluster < 1 || cluster > n_of_clusters){
            LOG(ERROR, "File %s should contain only integers >=%d and <=%d; '%s' found...",
                       in_assigned.c_str(), 1, n_of_clusters, str.c_str());
          }
          clusters->at(--cluster).insert(gene_no); 
          assigned->at(gene_no) = cluster;
          gene_no++;
        }
        delete l;
      }
      if(gene_no != nverts)
  LOG(ERROR, "File %s should contain %d number of intergers, which is a number of clustered genes",
              in_assigned.c_str(), nverts);
      delete deq;
    }
  }
  
  int kmeans_choose_best(Graph * g, clusters_t * clusters, int from, scores_t * distances = 0){
    int allocated = 0;
    if(!distances){ allocated = 1; distances = new scores_t(clusters->size()); }

    for(int i = 0, iend = clusters->size(); i < iend; ++i)
      kmeans_calculate_distance(g, &clusters->at(i), from, &distances->at(i));
    
    double best_val = std::numeric_limits<double>::max(); 
    int best_i = 0; 
    for(int i = 0, iend = clusters->size(); i < iend; ++i)
      if(distances->at(i) < best_val) best_val = distances->at(best_i = i);

    if(allocated) delete distances;
    return best_i;
  }

  void kmeans_assign_rest(Graph * g, clusters_t * clusters, assigned_t * assigned){
    for(int i = 0, iend = assigned->size(); i < iend; ++i){
      int & to = assigned->at(i);
      if(to == -1){
        to = kmeans_choose_best(g, clusters, i);
        clusters->at(to).insert(i);
      }
    }
  }

  void kmeans_fillup_assigned(Graph * g, clusters_t * clusters, assigned_t * assigned){
    int n_of_clusters = clusters->size();
    int nverts = clusters->size();
    foreach(int & i, *assigned) i = -1;
    for(int i = 0, v; i < n_of_clusters; ++i)
      foreach(int j, clusters->at(i)) assigned->at(j) = i;
  }

  double kmeans_sum(scores_t * scores){
    double ret = 0;
    foreach(double d, *scores) ret += d;
    return ret;
  }

  void res_to_assignment(Results * res, assigned_t * assignment){
    assert(assignment);
    assignment->resize(0);
    assignment->resize(res->get_no_vertices(), -1);
    for(int i = 0; i < res->get_no_results(); ++i){
      Results::Result & r = res->get(i);
      foreach(Results::ResultBinEl p, r.second)
        assignment->at(p.first) = i;
    }
  }

  Results * kmeans_lloyd(Graph * g, int n_of_clusters, int iterations, string in_assignment)
  {
    assert(n_of_clusters <= g->get_no_vertices());

    assigned_t assigned(g->get_no_vertices(), -1);
    clusters_t clusters(n_of_clusters);

    //these are actually distances. The score is assumed to be the higher, the better.
    scores_t scores(n_of_clusters, 0);
    int nverts = g->get_no_vertices();
    
    LOG(DEBUG, "Beginning finding optimal ;) %d clusters", n_of_clusters);
    kmeans_assign_seeds(g, &clusters, &assigned, in_assignment);
    if(verbosity >= DEBUG){
      std::string seeds; seeds.reserve(1000);
      for(int i = 0; i < n_of_clusters; ++i) 
        seeds += " " +
          boost::lexical_cast<std::string>(*clusters[i].begin());
      LOG(DEBUG, "Chosen seeds: %s", seeds.c_str());
    }
    kmeans_fillup_assigned(g, &clusters, &assigned);
    kmeans_assign_rest(g, &clusters, &assigned);
    
    kmeans_recalculate_sumdistances(g, &clusters, &scores);
    double sum2, sum = kmeans_sum(&scores);
    if(verbosity >= DEBUG){
      std::string clusts; clusts.reserve(1000);
      for(int i = 0; i < n_of_clusters; ++i) 
        clusts += " " + boost::lexical_cast<std::string>(clusters[i].size());
      LOG(DEBUG, "Initial clusters' distances: %g and sizes: %s", sum, 
                                                    clusts.c_str());
    }

    int it = 0;
    for(int i = 0; i != iterations; ++i){
      sum2 = sum;

      for(int i = 0; i < nverts; ++i){
        int & to = assigned[i];
        int to2 = kmeans_choose_best(g, &clusters, i);
        if(to != to2){
          double distance_to, distance_to2, sum_tmp, sum_tmp2;
          if(verbosity >= DEBUG){
            kmeans_calculate_distance(g, &clusters[to ], i, &distance_to );
            kmeans_calculate_distance(g, &clusters[to2], i, &distance_to2);

            kmeans_recalculate_sumdistances(g, &clusters, &scores);
            sum_tmp = kmeans_sum(&scores);
          }
          clusters[to].erase(i);
          clusters[to = to2].insert(i);
          if(verbosity >= DEBUG){
            kmeans_recalculate_sumdistances(g, &clusters, &scores);
            sum_tmp2 = kmeans_sum(&scores);
            LOG(DEBUG, "Changing %d from cluster %d (dist=%g), to %d (dist=%g); sumdistance, changed from %g to %g", 
                        i, to, distance_to, to2, distance_to2, sum_tmp, sum_tmp2);
          }
        }
      }

      kmeans_recalculate_sumdistances(g, &clusters, &scores);
      sum = kmeans_sum(&scores);

      if(verbosity >= DEBUG){
        std::string clusts; clusts.reserve(1000);
        for(int i = 0; i < n_of_clusters; ++i) 
          clusts += " " + boost::lexical_cast<std::string>(clusters[i].size());
        LOG(DEBUG, "It(%d) clusters' distances: %g and sizes: %s", it++, sum, 
                                                      clusts.c_str());
      }
      if(sum2 == sum) break;
    }
    
    Results * res = new Results(sum, g->get_no_vertices()); 
    res->set_names(&g->get_name(0), g->get_no_vertices(), 0);
    res->set_descs(&g->get_desc(0), g->get_no_vertices(), 0);
    deque<Results::Result> & deqres = res->get_deq();
    
    for(int i = 0, iend = clusters.size(); i < iend; ++i){
      deqres.push_back(Results::Result());
      Results::Result & r = deqres.back();
      r.first = scores[i];
      foreach(int j, clusters[i]) r.second[j] = 0;
    }
    return res;
  }


 
  InterfaceFunction{

    if(package->interface && package->interface->arg != InterfaceFlow::None)
      return 0;

    int n = 1;
    int n_of_clusters = 1;
    long iterations = -1;
    string out_assignment = "";
    string in_assignment = "";

    bool best = 0;
    int vector = 0;
    char str[100];
    sprintf(str, "%s:p[%d]", package->name.c_str(), package->packageNumber);
    timer::Timer t(str);
    t.start();
      ArgAtoms outputs;
      deque<allinone::Graph *> graphs;
      Interface * interface;
      long seed = -1;

      for(ArgAtoms::iterator it = package->argNode.begin(); 
            it != package->argNode.end(); ++it){
        if((*it)->argAtomType & (ArgAtom::LongSwitch | ArgAtom::ShortSwitch)){
          if((*it)->value == "-i" || (*it)->value == "--input"){
            *++it;
            switch((*it)->argAtomType){
            case ArgAtom::ToProcess:
            case ArgAtom::PackageNumber:
              getInterfaces(&graphs, (*it), InterfaceFlow::Graph);
              break;

            case ArgAtom::Arg:
              LOG(WARNING, "The package kmeans doesn't read files; use graph package instead");
            break;
            }
          }else if((*it)->value == "-o" || (*it)->value == "--output"){
            outputs.push_back(*++it);
          }else if((*it)->value == "-I" || (*it)->value == "--max-iterations"){
            if(!toInt((*++it)->value, &iterations, true))
              LOG(ERROR, "Cannot transform value to integer: '%s', aborting", 
                          (*++it)->value.c_str());
          }else if((*it)->value == "-a" || (*it)->value == "--initial-clusters"){
            in_assignment = (*++it)->value;
          }else if((*it)->value == "-A" || (*it)->value == "--save-clusters"){
            out_assignment = (*++it)->value;
          }else if((*it)->value == "-N" || (*it)->value == "--n-of-clusters"){
            long tmp_n_of_clusters = -1;
            if(!toInt((*++it)->value, &tmp_n_of_clusters, true))
              LOG(ERROR, "Cannot transform value to integer: '%s', aborting", 
                          (*++it)->value.c_str());
            else n_of_clusters = tmp_n_of_clusters;
          }else if((*it)->value == "-s" || (*it)->value == "--seed"){
            long tmp_seed = -1;
            if(!toInt((*++it)->value, &tmp_seed, true))
              LOG(ERROR, "Cannot transform value to integer: '%s', aborting", 
                          (*++it)->value.c_str());
            else seed = tmp_seed;
          }else if((*it)->value == "-n" || (*it)->value == "--number-of"){
            n = atoi((*++it)->value.c_str());
            if(n < 1){
              LOG(ERROR, "Wrong argument of \"--number-of\" switch: %s", 
                        (*it)->value.c_str());
            }
          }else if((*it)->value == "-b" || (*it)->value == "--best"){
            best = true;
          }else if((*it)->value == "-v" || (*it)->value == "--vector"){
            assert(!vector); vector = 1;
          }else if((*it)->value == "-V" || (*it)->value == "--r-vector"){
            assert(!vector); vector = 2;
          }
        }
      }

      if(graphs.size() != 1)
        LOG(ERROR, "Wrong number of input packages, 1 is required: ", 
          graphs.size());

      Graph * g = graphs.front();

      if(n_of_clusters == -1){
        n_of_clusters = std::min<int>(20, g->get_no_vertices());
        LOG(WARNING, "You didn't set the number of clusters."
                     " Setting default = %d;", n_of_clusters);
      }
      
      if(seed != -1){
        srand(seed);
      }else if(!seed_initialized){
        timeval tv; gettimeofday(&tv, 0);
        srand(tv.tv_usec);
      }

      std::vector<Results *> vr(vector == 2 && best ? 1 : n);
      assigned_t assigned;

      switch(vector){
        case 0: for(int i = 0; i < n; ++i) 
            vr[i] = kmeans_lloyd(g, n_of_clusters, iterations, in_assignment);
        break;
        case 1: for(int i = 0; i < n; ++i) 
            vr[i] = kmeans_euc::kmeans_lloyd(g, n_of_clusters, iterations, in_assignment);
        break;
        case 2: for(int i = 0; i < (best ? 1 : n); ++i) 
            vr[i] = kmeans_euc::kmeans_r(g, n_of_clusters, iterations, best ? n : 1, in_assignment);
        break;
      }

      std::sort(vr.begin(), vr.end(), allinone::Results_less(1));

      if(best || n == 1) {
        package->interface = vr[0];
        if(out_assignment != ""){
          res_to_assignment(vr[0], &assigned);
          kmeans_store_assignment(&assigned, out_assignment);
        }
      }
      else {
        Multiple * m = new Multiple();
        foreach(Results * i, vr) m->interfaces.push_back((Interface*)i);
        package->interface = m;
        if(out_assignment != ""){
          for(int i = 0; i < vr.size(); ++i){
            res_to_assignment(vr[i], &assigned);
            kmeans_store_assignment(&assigned, out_assignment + "_" + toString(i));
          }
        }
      }

      package->interface->readOnly=true;
      package->interface->freeAble=false;

    t.stop();
    return 0;
  }
}



