#include"config.h"
#include"graph/graph.hpp"
//#include"graph_interface.hpp"
#include"constants.h"
#include<stdio.h>
#include<deque>
#include<vector>
#include<set>
#include<list>
using namespace std;
#ifdef HAVE_TIMER_INTERFACE_H
#  include"timer/timer.h"  
#endif

namespace ppi {


  using allinone::ArgAtoms;
  using allinone::ArgAtom;
  using allinone::Packages;
  using allinone::Package;
  using allinone::Graph;
  
  void save_file(string name, Graph * g, char delimit){
    if(!g) return;
    string name2 = allinone::getWriteFilename(name);
    FILE * f = name2 == "-" ? stdout : fopen(name2.c_str(), "w");
    if(!f){
      LOG(WARNING, "Can't open file for write: %s, skipping", name2.c_str());
      return;
    }
    int columns = g->get_vector_length();
    int genes = g->get_no_vertices();

    for(int i = 0; i < genes; ++i){
      for(int j = 0; j <= i; ++j){
        if(g->get(i, j, 0) == 1)
          fprintf(f, "%s%cpp%c%s\n", g->get_name(i), delimit, delimit, g->get_name(j));
        else if(g->get(j, i, 0) == 1)
          fprintf(f, "%s%cpp%c%s\n", g->get_name(j), delimit, delimit, g->get_name(i));
      }
    }
    if(name2 != "-") fclose(f);
  }

  Graph * read_file(string name, char delimit, string datatype, string graphtype){
    deque<string> * deq = allinone::getStringsFromFile(name.c_str());
    list<string> * l = 0;
    if(!deq) LOG(ERROR, "Can't read file: %s", name.c_str());
    size_t lines = deq->size();
    if(lines < 2) LOG(ERROR, "File: %s has less than two lines", name.c_str());
    
    const char * tmpstr; 
    typedef pair<string, string> _2Str; 
    deque<_2Str> cons;
    set<string> genes;
    int size;

    foreach(string & str, *deq){
      if(l) { delete l; l = 0; }
      l = allinone::splitNStrip(str, delimit);
      size = l->size();
      if(!size || size == 1 && l->front() == "") continue;
      if(size != 3) {
        LOG(WARNING, "Wrong number of elemnts (%d) in the line: %s, should be: %d; skipping...",
            size, str.c_str(), 3);
        continue;
      }
      genes.insert(l->front());
      genes.insert(l->back());
      cons.push_back(_2Str(l->front(), l->back()));
    }

    Graph *g = graph::allocate_graph(false, true, 1, genes.size(), datatype, graphtype);
    int i = 0;
    foreach(const string & str, genes) 
      g->set_names(&(tmpstr = str.c_str()), 1, i++);
    
    foreach(_2Str & con, cons){
      g->set(con.first.c_str(), con.second.c_str(), 0, 1);
      g->set(con.second.c_str(), con.first.c_str(), 0, 1);
    }
    
    delete deq;
    return g;
  }


  InterfaceFunction{
    if(package->interface && package->interface->arg != InterfaceFlow::None)
      return 0;
    int n = 1;
    char str[100];
    sprintf(str, "%s:p[%d]", package->name.c_str(), package->packageNumber);
    timer::Timer t(str);
    t.start();

      char delimit = '\t';
      string datatype  = "";
      string graphtype = "";
      ArgAtoms outputs;
      deque<Graph *> graphs;
      Package * p;
      Interface * interface;

      for(ArgAtoms::iterator it = package->argNode.begin(); 
            it != package->argNode.end(); ++it){
        if((*it)->argAtomType & (ArgAtom::LongSwitch | ArgAtom::ShortSwitch)){
          if((*it)->value == "-i" || (*it)->value == "--input"){
            *++it;
            switch((*it)->argAtomType){
            case ArgAtom::ToProcess:
            case ArgAtom::PackageNumber:
              getInterfaces(&graphs, (*it), InterfaceFlow::Graph);
               break;
            case ArgAtom::Arg:
              graphs.push_back(read_file((*it)->value, delimit, datatype, graphtype));
            break;
            }
          }else if((*it)->value == "-o" || (*it)->value == "--output"){
            outputs.push_back(*++it);
          }else if((*it)->value == "-n" || (*it)->value == "--number-of"){
            n = atoi((*++it)->value.c_str());
            if(!(n >= 1)){
              LOG(ERROR, "Wrong argument of \"--number-of\" switch: %d", 
                        (*it)->value.c_str());
            }
          }else if((*it)->value == "-g" || (*it)->value == "--graph-type"){
            graphtype = (*++it)->value;
          }else if((*it)->value == "-t" || (*it)->value == "--data-type"){
            datatype = (*++it)->value;
          }else if((*it)->value == "-d"){
            delimit = allinone::parseDelimiter((*++it)->value);
          }
        }
      }

      Graph * graph;
      list<Interface *> interfaces;
      switch(graphs.size()){
      case 0:
        LOG(ERROR, "No input source specified for package[%d]: %s", 
          package->packageNumber, package->name.c_str());
      case 1:
        graph = graphs.front(); break;
        if((graphtype == "" || graphtype == graph->get_graphtype_s()) && 
           (datatype == "" || datatype == graph->get_datatype_s()));
        else 
          graph = graph::allocate_graph(graph, graph->get_no_vertices(), datatype, graphtype);
      default:
        LOG(ERROR, "Currently PPI package supports only one input graph.\n"
                   "You can merge them in Graph package.");
      }

    graph->readOnly = true;
    graph->freeAble = false;

    package->interface = graph;

    foreach(ArgAtom * o, outputs)
      save_file(o->value, graph, delimit);
    
    t.stop();
    return 0;
  }
}
