#include"config.h"
#include"constants.h"
#include<stdio.h>
#include<results.hpp>
#include<deque>
#include<vector>
#include<set>
#include<list>
using namespace std;
#ifdef HAVE_TIMER_INTERFACE_H
#  include"timer/timer.h"  
#endif

namespace results {

  using allinone::ArgAtoms;
  using allinone::ArgAtom;
  using allinone::Packages;
  using allinone::Package;
  using allinone::Results;
  
  void save_file_bynumbers(FILE *f, Results * r, char delimit){
    int genes = r->get_no_vertices ();
    int results = r->get_no_results();
    fprintf(f, "%d%c%d%c%g", genes, delimit, results, delimit, r->get_score());
    int count = 0;
    for(int i = 0; i < genes; ++i){
      if(!(count++ % 10)) fprintf(f, "\n%s", r->get_name(i));
      else fprintf(f, "%c%s", delimit, r->get_name(i));
    }
    fprintf(f, "\n\n");

    count=0;
    for(int i = 0; i < genes; ++i){
      const char *tmp = r->get_desc(i);
      if(!(count++ % 10)) fprintf(f, "\n%s", tmp?:"");
      else fprintf(f, "%c%s", delimit, tmp?:"");
    }
    fprintf(f, "\n\n");
    
    for(Results::iterator it=r->begin(), itend=r->end(); it != itend; it++){
      fprintf(f, "%.20g", it->first);
      typedef std::pair<int, _PFloat> Res;
      foreach(Res p, it->second){
        fprintf(f, "%c%d%c%.20g", delimit, p.first, delimit, p.second);
      }
      fprintf(f, "\n");
    }
  }
  
  void save_file_bynames(FILE *f, Results * r, char delimit){
    int genes = r->get_no_vertices ();
    int results = r->get_no_results();
    fprintf(f, "%d%c%d%c%g\n\n", genes, delimit, results, delimit, r->get_score());

    for(Results::iterator it=r->begin(), itend=r->end(); it != itend; it++){
      fprintf(f, "%.20g", it->first);
      typedef std::pair<int, _PFloat> Res;
      foreach(Res p, it->second){
        fprintf(f, "%c%s%c%.20g", delimit, r->get_name(p.first), delimit, p.second);
      }
      fprintf(f, "\n");
    }
  }

  void save_file(string name, Results * r, char delimit, string type){
     if(!r) return;
    string name2 = allinone::getWriteFilename(name);
    FILE * f = name2 == "-" ? stdout : fopen(name2.c_str(), "w");
    if(!f){
      LOG(WARNING, "Can't open file for write: %s, skipping", name2.c_str());
      return;
    }
    if(type == "bynames"){
      save_file_bynumbers(f, r, delimit);
    }else{
      save_file_bynames(f, r, delimit);
    }
  }

  Results * read_file(string name, char delimit){
    deque<string> * deq = allinone::getStringsFromFile(name.c_str());
    list<string> * l = 0;
    if(!deq) LOG(ERROR, "Can't read file: %s", name.c_str());
    size_t lines = deq->size();
    if(lines < 7) LOG(ERROR, "File: %s has less than 7 lines what is minimum of the graph file", name.c_str());
    
    const char * tmpstr; 
    long no_vertices = 0;
    long no_results = 0;
    double score = 0;
    const char * tmp;
    l = allinone::splitNStrip((*deq)[0], delimit);
    if(l->size() != 3 || !allinone::toInt(l->front(), &no_vertices) 
                      || !allinone::toInt(l->back(), &no_results)
                      || !allinone::toFloat(l->back(), &score))
      LOG(ERROR, "First decoded line: %s, should have 2 integers and 1 float.",
                   (*deq)[0].c_str());
    delete l;
    if(!(no_vertices > 0 && no_results > 0) || 
        no_vertices > 1000*1000 || no_results > 1000*1000){
      LOG(ERROR, "Wrong number of results/genes."
      " The number has to be bigger than 0 and the results object is limited to 1G results."
      "And is: %ld, %ld", no_vertices, no_results);
    }

    allinone::Results * r = new allinone::Results(score, no_vertices);
    
    long genes = 0;
    int line = 1; 
    
    allinone::skipEmpty(deq, line, delimit);
    for(l = allinone::splitNStrip((*deq)[line], delimit); 
          line < deq->size() && 
          !(l->size() == 0 || l->size() == 1 && l->front() == "") ; 
          l = allinone::splitNStrip((*deq)[++line], delimit)){
      foreach(string & str, *l){
        if(genes >= no_vertices) genes++; 
        else r->set_names(&(tmp = str.c_str()), 1, genes++);
      } delete l;
      if(line == deq->size() - 1) break;
    } 
    if(line != deq->size() - 1) delete l;
    if(genes > no_vertices) 
      LOG(WARNING, "To much genes names: %ld, expected: %ld", genes, no_vertices);
    if(genes < no_vertices) 
      LOG(ERROR, "To few genes names: %ld, expected: %ld", genes, no_vertices);
    if(deq->size() <= line) 
      LOG(ERROR, "File ended unexpectedly after gene names");

    genes = 0;
    allinone::skipEmpty(deq, line, delimit);
    for(l = allinone::splitNStrip((*deq)[line], delimit); 
          line < deq->size() && 
          !(l->size() == 0 || l->size() == 1 && l->front() == "") ; 
          l = allinone::splitNStrip((*deq)[++line], delimit)){
      foreach(string & str, *l){
        if(genes >= no_vertices) genes++; 
        else r->set_descs(&(tmp = str.c_str()), 1, genes++);
      } delete l;
      if(line == deq->size() - 1) break;
    } 
    if(line != deq->size() - 1) delete l;

    if(genes > no_vertices) 
      LOG(WARNING, "To much genes descriptions: %ld, expected: %ld, skipping...", 
        genes, no_vertices);
    if(deq->size() <= line) 
      LOG(ERROR, "File ended unexpectedly after genes descriptions");

    double val = 0;
    int results = 0;
    allinone::skipEmpty(deq, line, delimit);
    for(l = allinone::splitNStrip((*deq)[line], delimit); 
          line < deq->size() && 
          !(l->size() == 0 || l->size() == 1 && l->front() == "") ; 
          l = allinone::splitNStrip((*deq)[++line], delimit)){

      r->res.push_back(allinone::Results::Result());
      allinone::Results::Result & res = *r->rbegin();

      list<string>::iterator it = l->begin();
      if(allinone::toFloat(*it, &res.first))
        LOG(ERROR, "Can't decode results score: %s", it->c_str());
      int i = 0; 
      long tmpL;
      std::pair<int, _PFloat> pr;
      for(++it; it!=l->end(); ++it, ++i){
        if(!(i%2) && allinone::toInt(*it, &tmpL))
          pr.first=tmpL;
        else if((i%2) && allinone::toFloat(*it, &pr.second)){
          res.second.insert(pr);  
          ++results;
        }
        else LOG(ERROR, "Can't decode %s: %s", (i%2) ? "float" : "int", it->c_str());
      } delete l;
      if(line == deq->size() - 1) break;
    }
    if(line != deq->size() - 1) delete l;
    if(results > no_results)
      LOG(WARNING, "There are %ld results, expected: %ld ", 
            results, no_results);
    if(results < no_results)
      LOG(ERROR, "There are %ld results, expected: %ld ", 
            results, no_results);
    
    delete deq;
    return r;
  }


  InterfaceFunction{
    if(package->interface && package->interface->arg != InterfaceFlow::None)
      return 0;
    int n = 1;
    char str[100];
    sprintf(str, "%s:p[%d]", package->name.c_str(), package->packageNumber);
    timer::Timer t(str);
    t.start();

      char delimit = ';';
      ArgAtoms outputs;
      deque<Results *> resultss;
      Package * p;
      Interface * interface;
      string type = "bynumbers";

      for(ArgAtoms::iterator it = package->argNode.begin(); 
            it != package->argNode.end(); ++it){
        if((*it)->argAtomType & (ArgAtom::LongSwitch | ArgAtom::ShortSwitch)){
          if((*it)->value == "-i" || (*it)->value == "--input"){
            *++it;
            switch((*it)->argAtomType){
            case ArgAtom::ToProcess:
            case ArgAtom::PackageNumber:
              getInterfaces(&resultss, (*it), InterfaceFlow::Results);
              break;
            case ArgAtom::Arg:
              resultss.push_back(read_file((*it)->value, delimit));
            break;
            }
          }else if((*it)->value == "-o" || (*it)->value == "--output"){
            outputs.push_back(*++it);
          }else if((*it)->value == "-t" || (*it)->value == "--type"){
            type = (*++it)->value;
          }
          else if((*it)->value == "-n" || (*it)->value == "--number-of"){
            n = atoi((*++it)->value.c_str());
            if(!(n >= 1)){
              LOG(ERROR, "Wrong argument of \"--number-of\" switch: %d", 
                        (*it)->value.c_str());
            }
          }else if((*it)->value == "-d"){
            delimit = allinone::parseDelimiter((*++it)->value);
          }
        }
      }

      Results * results;
      list<Interface *> interfaces;
      switch(resultss.size()){
      case 0:
        LOG(ERROR, "No input source specified for package[%d]: %s", 
          package->packageNumber, package->name.c_str());
      case 1:
        results = resultss.front(); break;
      default:
        LOG(ERROR, "Currently Results package supports only one input graph.\n");
      }

      results->readOnly = true;
      results->freeAble = false;
     
      package->interface = results;
     
      foreach(ArgAtom * o, outputs)
        save_file(o->value, results, delimit, type);
    t.stop();

    return 0;
  }
}
