#include"timer.h"
#include "cycle.h"
#include "constants.h"
#include <unistd.h>
#include <sys/time.h>
#include <stdio.h>
#include <assert.h>

namespace timer {

    static TimerSet * timerSet = 0;;

    TimerSet::TimerSet(){
      computeTicksPerSecond();
    }

    TimerSet * TimerSet::getInstance(){
      if(!timerSet) timerSet = new TimerSet();
      return timerSet;
    }

    void TimerSet::addTime(string const & name, const long & start, const long & stop){
      //TODO consumes memory as hell
//      timers[name].push_back(pair<long, long>(start, stop));
      summary[name] += stop - start;
    }
  
    double TimerSet::computeTicksPerSecond(){
      getticks(); getticks(); getticks();
      
      long ticks_start = getticks();
      double start = getTime();
      double end = start + 0.1;
      for(;getTime() < end;);
      long ticks_end = getticks();
      ticksPerSecond = (ticks_end - ticks_start) / 0.1;
      //printf("ticksPerSecond: %lf\n", ticksPerSecond);
      return ticksPerSecond;
    }
    
    double TimerSet::getTime(){
      timeval tv;
      gettimeofday(&tv, 0);
      return tv.tv_sec + tv.tv_usec / 1000000.0;
    }

    void TimerSet::print(FILE * f){
      pair<string, long> t;
      foreach(t, summary)
        fprintf(f, "%s: %fs\n", t.first.c_str(), t.second / ticksPerSecond) ;
    }

    Timer::Timer(char const * name) : name(string(name)), running(0){ }

    Timer::~Timer(){}
  
    void Timer::start(){
      assert(!running);
      started = getticks();
      running=true;
    }

    double Timer::stop(){
      assert(running);
      running=false;
      stopped = getticks();
      TimerSet::getInstance()->addTime(name, started, stopped);
      return (stopped - started) / TimerSet::getInstance()->getTicksPerSecond();
    }
}




