#ifndef GENES_HPP_
#define GENES_HPP_

#include "config.h"
#include "constants.h"
#include <string>
#include <stdlib.h>

namespace allinone {

  class Genes : public Interface{

    protected:
    vector<const char *> vertices_names;
    vector<const char *> vertices_descs;
    map<const char *, int, allinone::charptr_less> vertices_names_to_i;

    virtual void init (long no_vertices){
      vertices_names.resize(no_vertices);
      vertices_descs.resize(no_vertices);
      for(int i = 0; i < vertices_names.size(); ++i){
        vertices_descs[i] = 0;
        vertices_names[i] = 0;
        vertices_names_to_i.clear();
      }
    }

    public:

    Genes(long no_vertices, const Interface * interface) : 
        Interface(interface) {
      init(no_vertices); 
    }

    Genes(long no_vertices, 
        InterfaceFlow::InterfaceArg arg = InterfaceFlow::None) : 
        Interface(arg){
      init(no_vertices);
    }

    virtual ~Genes(){
      foreach(const char * i, vertices_names) free((void *)i);
      foreach(const char * i, vertices_descs) free((void *)i);
    }

    void set_names(const char ** names, long no, long offset = 0){
      const char * tmp;
      map<const char *, int, allinone::charptr_less>::iterator it;
      for(long i = offset, iend = no + offset; i < iend; i++){
        if((tmp = vertices_names[i]) != 0 && 
           (it = vertices_names_to_i.find(tmp)) != vertices_names_to_i.end()){
          vertices_names_to_i.erase(it);
          free((void *)tmp);
        }
        vertices_names[i] = tmp = strdup(names[i -offset]);
        vertices_names_to_i[tmp] = i;
      }
    }

    void set_descs(const char ** names, long no, long offset = 0){
      const char * tmp;
      map<const char *, int, allinone::charptr_less>::iterator it;
      for(long i = offset, iend = no + offset; i < iend; i++){
        if((tmp = vertices_descs[i]) != 0)
          free((void *)tmp);
        if(names[i - offset]) vertices_descs[i] = strdup(names[i - offset]);
        else vertices_descs[i] = 0;
      }
    }

    const char * & get_name(int vertex) { return vertices_names[vertex]; }

    const char * & get_desc(int vertex) { return vertices_descs[vertex]; }

    int get_no(const char * name){
      map<const char *, int, allinone::charptr_less>::iterator it = 
        vertices_names_to_i.find(name);
      if(it != vertices_names_to_i.end())
        return it->second;
      return 0;
    }

    virtual long get_no_vertices()   const = 0;

  };

}
#endif /*GENES_HPP_*/

